"""Tests for the Sphinx builder."""

from collections.abc import Callable
from importlib.metadata import version
from pathlib import Path

import docutils.utils
import pytest
from sphinx.errors import ExtensionError
from sphinx.testing.util import SphinxTestApp

import sphinx_notion


def test_meta(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Builder metadata and setup returns expected values for Sphinx
    integration.
    """
    builder_cls = sphinx_notion.NotionBuilder
    assert builder_cls.name == "notion"
    assert builder_cls.out_suffix == ".json"

    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()
    app = make_app(srcdir=srcdir)
    setup_result = sphinx_notion.setup(app=app)
    pkg_version = version(distribution_name="sphinx-notionbuilder")
    assert setup_result == {
        "parallel_read_safe": True,
        "version": pkg_version,
    }

    builder = builder_cls(app=app, env=app.env)
    document = docutils.utils.new_document(source_path=".")
    translator = sphinx_notion.NotionTranslator(
        document=document, builder=builder
    )
    translator.depart_document(node=document)
    assert translator.body == "[]"


def test_notion_publish_config_defaults(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """Default configuration values are set correctly."""
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()
    app = make_app(
        srcdir=srcdir,
        confoverrides={"extensions": ["sphinx_notion"]},
    )

    assert app.config.notion_publish is False
    assert app.config.notion_parent_page_id is None
    assert app.config.notion_parent_database_id is None
    assert app.config.notion_page_title is None
    assert app.config.notion_page_icon is None
    assert app.config.notion_page_cover_url is None
    assert app.config.notion_cancel_on_discussion is False


def test_notion_publish_requires_parent(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """Enabling publish without parent raises error."""
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()

    with pytest.raises(ExtensionError, match="neither notion_parent_page_id"):
        make_app(
            srcdir=srcdir,
            confoverrides={
                "extensions": ["sphinx_notion"],
                "notion_publish": True,
                "notion_page_title": "Test Page",
            },
        )


def test_notion_publish_mutually_exclusive_parents(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """Setting both parent_page_id and parent_database_id raises error."""
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()

    with pytest.raises(ExtensionError, match="mutually exclusive"):
        make_app(
            srcdir=srcdir,
            confoverrides={
                "extensions": ["sphinx_notion"],
                "notion_publish": True,
                "notion_parent_page_id": "abc123",
                "notion_parent_database_id": "def456",
                "notion_page_title": "Test Page",
            },
        )


def test_notion_publish_requires_title(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """Enabling publish without title raises error."""
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()

    with pytest.raises(ExtensionError, match="notion_page_title is not set"):
        make_app(
            srcdir=srcdir,
            confoverrides={
                "extensions": ["sphinx_notion"],
                "notion_publish": True,
                "notion_parent_page_id": "abc123",
            },
        )


def test_notion_publish_valid_config_with_page_id(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """Valid configuration with parent_page_id passes validation."""
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()

    app = make_app(
        srcdir=srcdir,
        confoverrides={
            "extensions": ["sphinx_notion"],
            "notion_publish": True,
            "notion_parent_page_id": "abc123",
            "notion_page_title": "Test Page",
        },
    )

    assert app.config.notion_publish is True
    assert app.config.notion_parent_page_id == "abc123"
    assert app.config.notion_page_title == "Test Page"


def test_notion_publish_valid_config_with_database_id(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """Valid configuration with parent_database_id passes validation."""
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()

    app = make_app(
        srcdir=srcdir,
        confoverrides={
            "extensions": ["sphinx_notion"],
            "notion_publish": True,
            "notion_parent_database_id": "def456",
            "notion_page_title": "Test Page",
            "notion_page_icon": "📚",
            "notion_page_cover_url": "https://example.com/cover.jpg",
            "notion_cancel_on_discussion": True,
        },
    )

    assert app.config.notion_publish is True
    assert app.config.notion_parent_database_id == "def456"
    assert app.config.notion_page_title == "Test Page"
    assert app.config.notion_page_icon == "📚"
    assert app.config.notion_page_cover_url == "https://example.com/cover.jpg"
    assert app.config.notion_cancel_on_discussion is True


def test_notion_publish_disabled_skips_validation(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """When publish is disabled, validation is skipped."""
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()

    app = make_app(
        srcdir=srcdir,
        confoverrides={
            "extensions": ["sphinx_notion"],
            "notion_publish": False,
        },
    )

    assert app.config.notion_publish is False



"""Configuration for ``pytest``."""

import json
import os
from collections.abc import Iterator
from unittest.mock import patch
from uuid import uuid4

import pytest
import respx
from beartype import beartype
from ultimate_notion import Session
from wiremock_mock import add_wiremock_to_respx

pytest_plugins = "sphinx.testing.fixtures"  # pylint: disable=invalid-name


@pytest.fixture(name="respx_mock", scope="module")
def fixture_respx_mock(
    *,
    request: pytest.FixtureRequest,
) -> Iterator[respx.MockRouter]:
    """Provide a respx mock router loaded with WireMock stubs."""
    mappings_path = (
        request.config.rootpath
        / "tests"
        / "notion_sandbox"
        / "notion-wiremock-stubs.json"
    )
    assert mappings_path.is_file()

    with mappings_path.open(encoding="utf-8") as mappings_file:
        stubs = json.load(fp=mappings_file)

    mock = respx.MockRouter(assert_all_called=False)
    add_wiremock_to_respx(
        mock_obj=mock,
        stubs=stubs,
        base_url="https://mock.notion.test",
    )
    mock.start()
    try:
        yield mock
    finally:
        mock.stop()


@pytest.fixture(name="mock_api_base_url", scope="module")
def fixture_mock_api_base_url_fixture(
    *,
    respx_mock: respx.MockRouter,
) -> str:
    """Provide a prepared mock service base URL."""
    del respx_mock
    return "https://mock.notion.test"


@pytest.fixture(name="notion_token")
def fixture_notion_token() -> Iterator[str]:
    """Provide a token in environment variables for Ultimate Notion."""
    token = uuid4().hex
    with patch.dict(in_dict=os.environ, values={"NOTION_TOKEN": token}):
        yield token


@pytest.fixture(name="notion_session")
def fixture_notion_session_fixture(
    *,
    mock_api_base_url: str,
    notion_token: str,
) -> Iterator[Session]:
    """Provide an `ultimate_notion` session wired to the mock API."""
    del notion_token
    session = Session(base_url=mock_api_base_url)
    try:
        yield session
    finally:
        session.close()


@pytest.fixture(name="parent_page_id")
def fixture_parent_page_id() -> str:
    """The page ID used by the mock API fixtures."""
    return "59833787-2cf9-4fdf-8782-e53db20768a5"


@beartype
def pytest_collection_modifyitems(items: list[pytest.Item]) -> None:
    """
    Apply the ``beartype`` decorator to all collected test
    functions.
    """
    for item in items:
        # All our tests are functions, for now
        assert isinstance(item, pytest.Function)
        item.obj = beartype(obj=item.obj)



"""Tests for the upload script."""

import json
from pathlib import Path
from typing import Any

from click.testing import CliRunner, Result
from pytest_regressions.file_regression import FileRegressionFixture
from ultimate_notion.blocks import (
    Paragraph as UnoParagraph,
)
from ultimate_notion.rich_text import text

from _notion_scripts.upload import main  # pylint: disable=import-private-name


def _write_blocks_file(
    *,
    tmp_path: Path,
    block_dicts: list[dict[str, Any]],
) -> Path:
    """Write block payload JSON to a temporary file."""
    blocks_file = tmp_path / "blocks.json"
    blocks_file.write_text(
        data=json.dumps(obj=block_dicts),
        encoding="utf-8",
    )
    return blocks_file


def _invoke_upload(
    *,
    blocks_file: Path,
    parent_page_id: str,
    mock_api_base_url: str,
    cancel_on_discussion: bool = False,
) -> Result:
    """Invoke the upload CLI against the mock Notion API."""
    runner = CliRunner()
    arguments = [
        "--file",
        str(object=blocks_file),
        "--parent-page-id",
        parent_page_id,
        "--title",
        "Upload Title",
        "--notion-api-base-url",
        mock_api_base_url,
    ]
    if cancel_on_discussion:
        arguments.append("--cancel-on-discussion")
    return runner.invoke(
        cli=main,
        args=arguments,
        catch_exceptions=False,
    )


def _paragraph_blocks(
    *,
    text_content: str,
) -> list[dict[str, Any]]:
    """Create serialized paragraph blocks."""
    return [
        UnoParagraph(
            text=text(text=text_content),
        ).obj_ref.serialize_for_api()
    ]


def test_help(file_regression: FileRegressionFixture) -> None:
    """Expected help text is shown.

    This help text is defined in files.
    To update these files, run ``pytest`` with the ``--regen-all`` flag.
    """
    runner = CliRunner()
    arguments = ["--help"]
    result = runner.invoke(
        cli=main,
        args=arguments,
        catch_exceptions=False,
        color=True,
    )
    assert result.exit_code == 0, (result.stdout, result.stderr)
    file_regression.check(contents=result.output)


def test_upload_success(
    *,
    mock_api_base_url: str,
    notion_token: str,
    parent_page_id: str,
    tmp_path: Path,
) -> None:
    """Uploading through the CLI reports success."""
    assert notion_token
    blocks_file = _write_blocks_file(
        tmp_path=tmp_path,
        block_dicts=_paragraph_blocks(
            text_content="Hello from WireMock upload test",
        ),
    )

    result = _invoke_upload(
        blocks_file=blocks_file,
        parent_page_id=parent_page_id,
        mock_api_base_url=mock_api_base_url,
    )

    assert result.exit_code == 0, (result.stdout, result.stderr)
    assert result.output == (
        "Uploaded page: 'Upload Title' "
        "(https://www.notion.so/Upload-Title-59833787)\n"
    )


def test_upload_page_has_subpages_error(
    *,
    mock_api_base_url: str,
    notion_token: str,
    tmp_path: Path,
) -> None:
    """A useful error is shown when the target page has subpages."""
    assert notion_token
    blocks_file = _write_blocks_file(
        tmp_path=tmp_path,
        block_dicts=[],
    )

    result = _invoke_upload(
        blocks_file=blocks_file,
        parent_page_id="aaaa0000-0000-0000-0000-000000000001",
        mock_api_base_url=mock_api_base_url,
    )

    assert result.exit_code == 1
    assert "This page has subpages." in result.output


def test_upload_page_has_databases_error(
    *,
    mock_api_base_url: str,
    notion_token: str,
    tmp_path: Path,
) -> None:
    """A useful error is shown when the target page has databases."""
    assert notion_token
    blocks_file = _write_blocks_file(
        tmp_path=tmp_path,
        block_dicts=[],
    )

    result = _invoke_upload(
        blocks_file=blocks_file,
        parent_page_id="bbbb0000-0000-0000-0000-000000000001",
        mock_api_base_url=mock_api_base_url,
    )

    assert result.exit_code == 1
    assert "This page has databases." in result.output


def test_upload_discussions_exist_error(
    *,
    mock_api_base_url: str,
    notion_token: str,
    tmp_path: Path,
) -> None:
    """The discussion error message is forwarded to the CLI."""
    assert notion_token
    blocks_file = _write_blocks_file(
        tmp_path=tmp_path,
        block_dicts=_paragraph_blocks(
            text_content="Different content triggers sync",
        ),
    )

    result = _invoke_upload(
        blocks_file=blocks_file,
        parent_page_id="cccc0000-0000-0000-0000-000000000001",
        mock_api_base_url=mock_api_base_url,
        cancel_on_discussion=True,
    )

    assert result.exit_code == 1
    assert "discussion" in result.output.lower()



"""Tests for the _publish_to_notion event callback."""

from collections.abc import Callable
from pathlib import Path

import pytest
import respx
from sphinx.errors import ExtensionError
from sphinx.testing.util import SphinxTestApp

from sphinx_notion._upload import PageHasSubpagesError
from tests._wiremock import (  # pyrefly: ignore[missing-import]
    count_mock_requests,
)


def test_publish_skips_on_exception(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """_publish_to_notion returns immediately when a build exception is
    given.
    """
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()
    app = make_app(
        srcdir=srcdir,
        confoverrides={"extensions": ["sphinx_notion"]},
    )
    warnings_before = app.warning.getvalue()
    app.emit("build-finished", RuntimeError("build failed"))
    assert app.warning.getvalue() == warnings_before


def test_publish_skips_when_disabled(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """_publish_to_notion returns immediately when notion_publish is False."""
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()
    (srcdir / "index.rst").write_text(data="Test\n====\n", encoding="utf-8")
    app = make_app(
        buildername="notion",
        srcdir=srcdir,
        confoverrides={
            "extensions": ["sphinx_notion"],
            "notion_publish": False,
        },
    )
    app.build()
    assert app.statuscode == 0


def test_publish_skips_when_not_notion_builder(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """_publish_to_notion returns when the active builder is not
    notion.
    """
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()
    (srcdir / "index.rst").write_text(data="Test\n====\n", encoding="utf-8")
    app = make_app(
        srcdir=srcdir,
        confoverrides={
            "extensions": ["sphinx_notion"],
            "notion_publish": True,
            "notion_parent_page_id": "abc123",
            "notion_page_title": "Test",
        },
    )
    app.build()
    assert app.statuscode == 0


def test_publish_skips_when_no_output_file(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """_publish_to_notion warns and returns when ``index.json`` does not
    exist.
    """
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()
    app = make_app(
        buildername="notion",
        srcdir=srcdir,
        confoverrides={
            "extensions": ["sphinx_notion"],
            "notion_publish": True,
            "notion_parent_page_id": "abc123",
            "notion_page_title": "Test",
        },
    )
    app.emit("build-finished", None)
    assert "index.json found" in app.warning.getvalue()


def test_publish_success(
    *,
    make_app: Callable[..., SphinxTestApp],
    mock_api_base_url: str,
    respx_mock: respx.MockRouter,
    notion_token: str,
    parent_page_id: str,
    tmp_path: Path,
) -> None:
    """A Notion build with notion_publish=True uploads to the mock Notion
    API.
    """
    del notion_token
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()
    (srcdir / "index.rst").write_text(
        data="Hello from publish test.\n",
        encoding="utf-8",
    )
    app = make_app(
        buildername="notion",
        srcdir=srcdir,
        confoverrides={
            "extensions": ["sphinx_notion"],
            "notion_publish": True,
            "notion_parent_page_id": parent_page_id,
            "notion_page_title": "Upload Title",
            "notion_api_base_url": mock_api_base_url,
        },
    )
    append_url_path = f"/v1/blocks/{parent_page_id}/children"
    before_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path=append_url_path,
    )
    app.build()
    assert app.statuscode == 0
    assert (
        count_mock_requests(
            mock=respx_mock,
            method="PATCH",
            url_path=append_url_path,
        )
        == before_count + 1
    )


def test_publish_propagates_error(
    *,
    make_app: Callable[..., SphinxTestApp],
    mock_api_base_url: str,
    notion_token: str,
    tmp_path: Path,
) -> None:
    """A Notion build with a failing upload propagates the error."""
    del notion_token
    srcdir = tmp_path / "src"
    srcdir.mkdir()
    (srcdir / "conf.py").touch()
    (srcdir / "index.rst").write_text(data="", encoding="utf-8")
    app = make_app(
        buildername="notion",
        srcdir=srcdir,
        confoverrides={
            "extensions": ["sphinx_notion"],
            "notion_publish": True,
            "notion_parent_page_id": "aaaa0000-0000-0000-0000-000000000001",
            "notion_page_title": "Upload Title",
            "notion_api_base_url": mock_api_base_url,
        },
    )
    with pytest.raises(expected_exception=ExtensionError) as exc_info:
        app.build()
    assert isinstance(exc_info.value.__cause__, PageHasSubpagesError)



"""WireMock test helpers."""

from __future__ import annotations

from typing import TYPE_CHECKING

from beartype import beartype

if TYPE_CHECKING:
    import respx
    from respx.models import Call


@beartype
def count_mock_requests(
    *,
    mock: respx.MockRouter,
    method: str,
    url_path: str,
) -> int:
    """Count matching requests captured by the respx mock."""
    calls: list[Call] = list(mock.calls)
    count = 0
    for call in calls:
        if call.request.method == method and call.request.url.path == url_path:
            count += 1
    return count



"""Integration test for upload synchronization against a mock API."""

from pathlib import Path

import pytest
import respx
from ultimate_notion import ExternalFile, Session
from ultimate_notion.blocks import (
    BulletedItem,
    Divider,
)
from ultimate_notion.blocks import Image as UnoImage
from ultimate_notion.blocks import (
    Paragraph as UnoParagraph,
)
from ultimate_notion.obj_api.objects import UploadedFile as UnoUploadedFile
from ultimate_notion.rich_text import text

import sphinx_notion._upload as notion_upload
from sphinx_notion._upload import (
    DiscussionsExistError,
    PageHasDatabasesError,
    PageHasSubpagesError,
)
from tests._wiremock import (
    count_mock_requests,
)


def _file_upload_create_count(*, mock: respx.MockRouter) -> int:
    """Count calls to file-upload creation endpoint."""
    return count_mock_requests(
        mock=mock,
        method="POST",
        url_path="/v1/file_uploads",
    )


def test_upload_to_notion_with_wiremock(
    *,
    notion_session: Session,
    parent_page_id: str,
) -> None:
    """It is possible to upload a page with the mock API."""
    page = notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoParagraph(text=text(text="Hello from WireMock upload test"))
        ],
        parent_page_id=parent_page_id,
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )

    assert page.title == "Upload Title"
    assert page.url == "https://www.notion.so/Upload-Title-59833787"
    assert str(object=page.id) == parent_page_id
    assert len(page.blocks) == 1
    assert isinstance(page.blocks[0], UnoParagraph)
    assert page.blocks[0].rich_text == "Hello from WireMock upload test"


def test_upload_deletes_and_replaces_changed_blocks(
    *,
    respx_mock: respx.MockRouter,
    notion_session: Session,
    parent_page_id: str,
) -> None:
    """Changed content triggers block deletion and re-upload."""
    before_delete_count = count_mock_requests(
        mock=respx_mock,
        method="DELETE",
        url_path="/v1/blocks/c02fc1d3-db8b-45c5-a222-27595b15aea7",
    )
    before_append_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path=f"/v1/blocks/{parent_page_id}/children",
    )
    page = notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoParagraph(text=text(text="Different content triggers sync"))
        ],
        parent_page_id=parent_page_id,
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=True,
    )
    after_delete_count = count_mock_requests(
        mock=respx_mock,
        method="DELETE",
        url_path="/v1/blocks/c02fc1d3-db8b-45c5-a222-27595b15aea7",
    )
    after_append_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path=f"/v1/blocks/{parent_page_id}/children",
    )

    assert page.title == "Upload Title"
    assert str(object=page.id) == parent_page_id
    assert after_delete_count == before_delete_count + 1
    assert after_append_count == before_append_count + 1


def test_upload_with_icon(
    *,
    notion_session: Session,
    parent_page_id: str,
) -> None:
    """It is possible to upload a page with an emoji icon."""
    page = notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoParagraph(text=text(text="Hello from WireMock upload test"))
        ],
        parent_page_id=parent_page_id,
        parent_database_id=None,
        title="Upload Title",
        icon="\N{MEMO}",
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )

    assert page.title == "Upload Title"
    assert str(object=page.id) == parent_page_id
    assert page.icon == "\N{MEMO}"


def test_upload_with_cover_url(
    *,
    notion_session: Session,
    parent_page_id: str,
) -> None:
    """It is possible to upload a page with a cover URL."""
    page = notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoParagraph(text=text(text="Hello from WireMock upload test"))
        ],
        parent_page_id=parent_page_id,
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url="https://example.com/cover.png",
        cancel_on_discussion=False,
    )

    assert page.title == "Upload Title"
    assert str(object=page.id) == parent_page_id
    assert isinstance(page.cover, ExternalFile)
    assert page.cover.url == "https://example.com/cover.png"


def test_upload_page_has_subpages_error(
    notion_session: Session,
) -> None:
    """PageHasSubpagesError raised when the target page has subpages."""
    with pytest.raises(expected_exception=PageHasSubpagesError):
        notion_upload.upload_to_notion(
            session=notion_session,
            blocks=[],
            parent_page_id="aaaa0000-0000-0000-0000-000000000001",
            parent_database_id=None,
            title="Upload Title",
            icon=None,
            cover_path=None,
            cover_url=None,
            cancel_on_discussion=False,
        )


def test_upload_page_has_databases_error(
    notion_session: Session,
) -> None:
    """PageHasDatabasesError raised when the target page has databases."""
    with pytest.raises(expected_exception=PageHasDatabasesError):
        notion_upload.upload_to_notion(
            session=notion_session,
            blocks=[],
            parent_page_id="bbbb0000-0000-0000-0000-000000000001",
            parent_database_id=None,
            title="Upload Title",
            icon=None,
            cover_path=None,
            cover_url=None,
            cancel_on_discussion=False,
        )


def test_upload_discussions_exist_error(
    notion_session: Session,
) -> None:
    """DiscussionsExistError raised when blocks to delete have discussions."""
    with pytest.raises(
        expected_exception=DiscussionsExistError,
        match=r"1 block.*1 discussion",
    ):
        notion_upload.upload_to_notion(
            session=notion_session,
            blocks=[
                UnoParagraph(
                    text=text(text="Different content triggers sync"),
                ),
            ],
            parent_page_id="cccc0000-0000-0000-0000-000000000001",
            parent_database_id=None,
            title="Upload Title",
            icon=None,
            cover_path=None,
            cover_url=None,
            cancel_on_discussion=True,
        )


def test_upload_with_database_parent(
    *,
    notion_session: Session,
    respx_mock: respx.MockRouter,
) -> None:
    """It is possible to upload a page to a database."""
    parent_database_id = "db000000-0000-0000-0000-000000000001"
    query_url_path = f"/v1/databases/{parent_database_id}/query"

    before_count = count_mock_requests(
        mock=respx_mock,
        method="POST",
        url_path=query_url_path,
    )

    page = notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoParagraph(text=text(text="Hello from Microcks upload test"))
        ],
        parent_page_id=None,
        parent_database_id=parent_database_id,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )

    after_count = count_mock_requests(
        mock=respx_mock,
        method="POST",
        url_path=query_url_path,
    )

    assert page.title == "Upload Title"
    assert after_count == before_count + 1


def test_upload_with_cover_path(
    *,
    respx_mock: respx.MockRouter,
    notion_session: Session,
    parent_page_id: str,
    tmp_path: Path,
) -> None:
    """It is possible to upload a page with a local cover file."""
    cover_file = tmp_path / "cover.png"
    cover_file.write_bytes(data=b"fake-png-data")
    before_upload_count = _file_upload_create_count(
        mock=respx_mock,
    )

    page = notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoParagraph(text=text(text="Hello from Microcks upload test"))
        ],
        parent_page_id=parent_page_id,
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=cover_file,
        cover_url=None,
        cancel_on_discussion=False,
    )
    after_upload_count = _file_upload_create_count(
        mock=respx_mock,
    )

    assert page.title == "Upload Title"
    assert str(object=page.id) == parent_page_id
    assert isinstance(page.cover, ExternalFile)
    assert page.cover.url == "https://example.com/cover.png"
    assert after_upload_count == before_upload_count + 1


def test_upload_with_file_block(
    *,
    notion_session: Session,
    parent_page_id: str,
    tmp_path: Path,
) -> None:
    """It is possible to upload a page with a file:// image block."""
    img_file = tmp_path / "test.png"
    img_file.write_bytes(data=b"fake-image-data")

    page = notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoImage(file=ExternalFile(url=img_file.as_uri())),
        ],
        parent_page_id=parent_page_id,
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )

    assert page.title == "Upload Title"
    assert str(object=page.id) == parent_page_id
    assert len(page.blocks) == 1
    uploaded_image_block = page.blocks[0]
    assert isinstance(uploaded_image_block, UnoImage)
    assert str(object=uploaded_image_block.id) == (
        "30f89f8f-57ff-4f6c-a13d-4720d0d4f123"
    )
    uploaded_image = uploaded_image_block.obj_ref.image
    assert uploaded_image is not None
    assert isinstance(uploaded_image, UnoUploadedFile)
    uploaded_image_file = uploaded_image.file_upload
    assert str(object=uploaded_image_file.id) == (
        "ff000000-0000-0000-0000-000000000001"
    )


def test_upload_with_nested_file_block(
    *,
    notion_session: Session,
    parent_page_id: str,
    tmp_path: Path,
) -> None:
    """Upload with a parent block containing a file child block."""
    img_file = tmp_path / "nested.png"
    img_file.write_bytes(data=b"fake-nested-image-data")

    parent_block = BulletedItem(text=text(text="Item with image"))
    parent_block.append(
        blocks=[UnoImage(file=ExternalFile(url=img_file.as_uri()))],
    )

    page = notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[parent_block],
        parent_page_id=parent_page_id,
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )

    assert page.title == "Upload Title"
    assert str(object=page.id) == parent_page_id
    assert len(page.blocks) == 1
    uploaded_parent_block = page.blocks[0]
    assert isinstance(uploaded_parent_block, BulletedItem)
    assert len(uploaded_parent_block.children) == 1
    uploaded_child_block = uploaded_parent_block.children[0]
    assert isinstance(uploaded_child_block, UnoImage)
    uploaded_child_image = uploaded_child_block.obj_ref.image
    assert uploaded_child_image is not None
    assert isinstance(uploaded_child_image, UnoUploadedFile)
    uploaded_child_file = uploaded_child_image.file_upload
    assert str(object=uploaded_child_file.id) == (
        "ff000000-0000-0000-0000-000000000001"
    )


def test_upload_prefix_suffix_matching(
    *,
    respx_mock: respx.MockRouter,
    notion_session: Session,
) -> None:
    """Prefix and suffix matching skips unchanged blocks."""
    before_delete_count = count_mock_requests(
        mock=respx_mock,
        method="DELETE",
        url_path="/v1/blocks/dddd0000-0000-0000-0000-000000000011",
    )
    before_append_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path="/v1/blocks/dddd0000-0000-0000-0000-000000000002/children",
    )
    notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoParagraph(text=text(text="same")),
            UnoParagraph(text=text(text="new")),
            Divider(),
        ],
        parent_page_id="dddd0000-0000-0000-0000-000000000001",
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )
    after_delete_count = count_mock_requests(
        mock=respx_mock,
        method="DELETE",
        url_path="/v1/blocks/dddd0000-0000-0000-0000-000000000011",
    )
    after_append_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path="/v1/blocks/dddd0000-0000-0000-0000-000000000002/children",
    )

    assert after_delete_count == before_delete_count + 1
    assert after_append_count == before_append_count + 1


def test_upload_file_block_name_mismatch(
    *,
    notion_session: Session,
    respx_mock: respx.MockRouter,
    tmp_path: Path,
) -> None:
    """File block with name mismatch triggers re-upload."""
    img_file = tmp_path / "different.png"
    img_file.write_bytes(data=b"image-data")

    before_upload_count = _file_upload_create_count(
        mock=respx_mock,
    )
    notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoImage(
                file=ExternalFile(
                    url=img_file.as_uri(),
                    name="different.png",
                ),
            ),
        ],
        parent_page_id="eeee0000-0000-0000-0000-000000000001",
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )
    after_upload_count = _file_upload_create_count(
        mock=respx_mock,
    )

    assert after_upload_count == before_upload_count + 1


def test_upload_file_block_caption_mismatch(
    *,
    notion_session: Session,
    respx_mock: respx.MockRouter,
    tmp_path: Path,
) -> None:
    """File block with caption mismatch triggers re-upload."""
    img_file = tmp_path / "test.png"
    img_file.write_bytes(data=b"image-data")

    before_upload_count = _file_upload_create_count(
        mock=respx_mock,
    )
    notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoImage(
                file=ExternalFile(url=img_file.as_uri()),
                caption="new caption",
            ),
        ],
        parent_page_id="eeee0000-0000-0000-0000-000000000001",
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )
    after_upload_count = _file_upload_create_count(
        mock=respx_mock,
    )

    assert after_upload_count == before_upload_count + 1


def test_upload_file_block_external_url(
    *,
    notion_session: Session,
    respx_mock: respx.MockRouter,
) -> None:
    """File block with external URL skips upload and compares directly."""
    before_upload_count = _file_upload_create_count(
        mock=respx_mock,
    )
    notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoImage(
                file=ExternalFile(
                    url="https://example.com/different.png",
                ),
            ),
        ],
        parent_page_id="eeee0000-0000-0000-0000-000000000001",
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )
    after_upload_count = _file_upload_create_count(
        mock=respx_mock,
    )

    assert after_upload_count == before_upload_count


def test_upload_file_block_existing_is_external(
    *,
    notion_session: Session,
    respx_mock: respx.MockRouter,
    tmp_path: Path,
) -> None:
    """File block with existing ExternalFile triggers re-upload."""
    img_file = tmp_path / "test.png"
    img_file.write_bytes(data=b"image-data")

    before_upload_count = _file_upload_create_count(
        mock=respx_mock,
    )
    notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[
            UnoImage(file=ExternalFile(url=img_file.as_uri())),
        ],
        parent_page_id="ffff0000-0000-0000-0000-000000000001",
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )
    after_upload_count = _file_upload_create_count(
        mock=respx_mock,
    )

    assert after_upload_count == before_upload_count + 1


def test_upload_matching_parent_blocks(
    *,
    respx_mock: respx.MockRouter,
    notion_session: Session,
) -> None:
    """Matching parent blocks with children are not re-uploaded."""
    local_block = BulletedItem(text=text(text="item"))
    local_block.append(blocks=[Divider()])

    before_delete_count = count_mock_requests(
        mock=respx_mock,
        method="DELETE",
        url_path="/v1/blocks/aabb0000-0000-0000-0000-000000000010",
    )
    before_append_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path="/v1/blocks/aabb0000-0000-0000-0000-000000000002/children",
    )
    notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[local_block],
        parent_page_id="aabb0000-0000-0000-0000-000000000001",
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )
    after_delete_count = count_mock_requests(
        mock=respx_mock,
        method="DELETE",
        url_path="/v1/blocks/aabb0000-0000-0000-0000-000000000010",
    )
    after_append_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path="/v1/blocks/aabb0000-0000-0000-0000-000000000002/children",
    )

    assert after_delete_count == before_delete_count
    assert after_append_count == before_append_count


def test_upload_parent_block_different_children_count(
    *,
    respx_mock: respx.MockRouter,
    notion_session: Session,
) -> None:
    """Parent block with different children count triggers re-upload."""
    local_block = BulletedItem(text=text(text="item"))
    local_block.append(blocks=[Divider(), Divider()])

    before_delete_count = count_mock_requests(
        mock=respx_mock,
        method="DELETE",
        url_path="/v1/blocks/aabb0000-0000-0000-0000-000000000010",
    )
    before_parent_append_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path="/v1/blocks/aabb0000-0000-0000-0000-000000000002/children",
    )
    before_child_append_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path="/v1/blocks/aabb0000-0000-0000-0000-000000000020/children",
    )
    notion_upload.upload_to_notion(
        session=notion_session,
        blocks=[local_block],
        parent_page_id="aabb0000-0000-0000-0000-000000000001",
        parent_database_id=None,
        title="Upload Title",
        icon=None,
        cover_path=None,
        cover_url=None,
        cancel_on_discussion=False,
    )
    after_delete_count = count_mock_requests(
        mock=respx_mock,
        method="DELETE",
        url_path="/v1/blocks/aabb0000-0000-0000-0000-000000000010",
    )
    after_parent_append_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path="/v1/blocks/aabb0000-0000-0000-0000-000000000002/children",
    )
    after_child_append_count = count_mock_requests(
        mock=respx_mock,
        method="PATCH",
        url_path="/v1/blocks/aabb0000-0000-0000-0000-000000000020/children",
    )

    assert after_delete_count == before_delete_count + 1
    assert after_parent_append_count == before_parent_append_count + 1
    assert after_child_append_count == before_child_append_count + 1
